from typing import Final, Callable
from google.cloud import pubsub_v1
from concurrent import futures
from concurrent.futures import wait
from .dassana_env import *
import json
import logging

publisher = None
topic_path = None
logger: Final = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

def publish_message(message, topic_name):
    global publisher
    global topic_path
    try:
        project_id = get_project_id()
        if not publisher:
            publisher = pubsub_v1.PublisherClient()
        publish_futures = []
        if not topic_path:
            topic_path = publisher.topic_path(project_id, topic_name)
        data = json.dumps(message)
        # When you publish a message, the client returns a future.
        publish_future = publisher.publish(topic_path, data.encode("utf-8"), timeout=180)
        publish_future.result(timeout=180)
        
        # Non-blocking. Publish failures are handled in the callback function.
        # publish_future.add_done_callback(get_callback(publish_future, data))

        # publish_futures.append(publish_future)
        # done, un_done = wait(publish_futures, return_when=futures.ALL_COMPLETED)
        # if len(un_done) > 0:
        #     logger.error(f"Failed To Publish all the messages to topic {topic_name}, Published {len(publish_futures)-len(un_done)}/{len(publish_futures)} Messages")

    except Exception as e:
        logger.error(f"Failed To Publish Message to topic {topic_name} Because of {e}")


import datetime
from uuid import uuid4

from .dassana_publisher import publish_message
from .dassana_env import *

from typing import Final
import logging, json
import dassana.dassana_exception as exc


logger: Final = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

tenant_id = get_tenant_id()
dassana_partner = get_partner()
dassana_partner_client_id = get_partner_client_id()
dassana_partner_tenant_id = get_partner_tenant_id()
config_id = get_ingestion_config_id()
app_id = get_app_id()

topic_name = None
if dassana_partner:
    topic_name = dassana_partner + "_log_event_topic"

scope_id_mapping = {
    "crowdstrike_edr": "detection",
    "crowdstrike_spotlight": "vulnerability",
    "tenable_vulnerability": "vulnerability",
    "snyk_vulnerability": "vulnerability",
    "prisma_cloud_cspm": "cspm",
    "prisma_cloud_cwpp": "vulnerability",
    "qualys_vulnerability": "vulnerability",
    "wiz_cwpp": "vulnerability",
    "wiz_cspm": "cspm",
    "prisma_cloud_security_group": "security-group",
    "prisma_cloud_instance": "instance",
    "carbon_black_vulnerability": "vulnerability",
    "ms_defender_endpoint_alert": "detection",
    "ms_defender_endpoint_vulnerability": "vulnerability"
}

def log(status=None, exception=None, locals={}, scope_id=None, metadata={}, job_id=None):
    state = build_state(scope_id, locals, job_id, status, exception)
    message = {}

    message["developerCtx"] = {}
    message["developerCtx"].update(state)
    message["developerCtx"].update(add_developer_context(metadata, state["status"], exception))

    message["customerCtx"] = {}
    message["customerCtx"].update(state)
    message["customerCtx"].update(add_customer_context(message["customerCtx"], exception))

    if state["status"] == "failed":
        logger.error(msg=json.dumps(message["developerCtx"]))
    else:
        logger.info(msg=json.dumps(message["developerCtx"]))
    
    message = message["customerCtx"]
    
    if dassana_partner:
        publish_message(message, topic_name)

def add_developer_context(metadata, status ,exception):
    state = {}
    state["tenantId"] = tenant_id
    if status == 'ready_for_loading' and metadata:
        state["source"] = {}
        state["source"]["pass"] = metadata["source"]["pass"]
        state["source"]["fail"] = metadata["source"]["fail"]
        state["source"]["debugLog"] = metadata["source"]["debug_log"]

    elif status == 'failed':
        state["errorDetails"] = {}

        if metadata:
            state["errorDetails"]["pass"] = metadata["pass"]
            state["errorDetails"]["fail"] = metadata["fail"]
            state["errorDetails"]["debugLog"] = metadata["debug_log"]

        if exception:
            if isinstance(exception, exc.DassanaException):
                state["errorDetails"]["errorCode"] = exception.error_type
                state["errorDetails"]["isInternal"] = exception.is_internal
                state["errorDetails"]["isAutoRecoverable"] = exception.is_auto_recoverable
                if exception.error_type == "internal_error":
                    if exception.error_details:
                        state["errorDetails"]["errorMessage"] = exception.error_details
                    else:
                        state["errorDetails"]["errorMessage"] = exception.message
                if isinstance(exception, exc.ApiError):
                    state["errorDetails"]["httpRequest"] = exception.http_request.__dict__
                    state["errorDetails"]["httpResonse"] = exception.http_response.__dict__ 
                return state
        
        if metadata and "error_code" in metadata:
            state["errorDetails"]["errorCode"] = metadata["error_code"]
        else:
            state["errorDetails"]["errorCode"] = "internal_error"

        if metadata and "is_internal" in metadata:
            state["errorDetails"]["isInternal"] = metadata["is_internal"]
        else:
            state["errorDetails"]["isInternal"] = True
        
        if metadata and "is_auto_recoverable" in metadata:
            state["errorDetails"]["isAutoRecoverable"] = metadata["is_auto_recoverable"]
        else:
            state["errorDetails"]["isAutoRecoverable"] = False

        if metadata and "error_message" in metadata:
            state["errorDetails"]["errorMessage"] = metadata["error_message"]
        else:
            state["errorDetails"]["errorMessage"] = "Unexpected error occurred while collecting data"
    return state
  
def add_customer_context(state, exception=None):


    state["status"] = "ok" if state.get("status") == "ready_for_loading" else state.get("status")
    if dassana_partner and dassana_partner_tenant_id:
        state["tenantId"] = dassana_partner_tenant_id
    elif dassana_partner and not dassana_partner_tenant_id:
        raise KeyError("DASSANA_PARTNER_TENANT_ID environment variable is not set. Review configuration.")


    if dassana_partner and dassana_partner_client_id:
        state["siteId"] = dassana_partner_client_id
    elif dassana_partner and not dassana_partner_client_id:
        raise KeyError("DASSANA_PARTNER_CLIENT_ID environment variable is not set. Review configuration.")

    if state["status"] == 'failed':
        state["errorDetails"] = {}
        if exception:
            if isinstance(exception, exc.DassanaException):
                if not exception.is_internal:
                    if isinstance(exception, exc.ApiError):
                        state["errorDetails"]["message"] = exception.message
                        state["errorDetails"]["httpRequest"] = exception.http_request.__dict__
                        state["errorDetails"]["httpResponse"] = exception.http_response.__dict__
                        return state
        state["errorDetails"]["message"] = "Job terminated due to internal error"
    return state

def build_state(scope_id, locals, job_id, status, exception):
    state = {}

    state["eventId"] = str(uuid4())
    state["timestamp"] = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
    state["connector"] = app_id
    if exception:
        state["status"] = "failed"
    elif not exception and not status:
        state["status"] = "in_progress"
    else:
        state["status"] = status
    
    state["level"] = "info" if state["status"] in ['ready_for_loading', 'in_progress'] else "error"

    if job_id:
        state["jobId"] = job_id

    if state["status"] == "failed":
        state["message"] = "failed to finish data collection"

    elif state["status"] == "in_progress": 
        state["message"] = "starting data collection"

    else:
        state["message"] = "successfully finished data collection"

    if 'config_id' in locals:
        state["connectionId"] = locals.get("config_id")
    else:
        state["connectionId"] = config_id
        
    if scope_id:
        state["scopeId"] = scope_id_mapping.get(scope_id, scope_id)

    return state


import timeit
import requests
from requests.models import Response
import logging
from tenacity import retry, stop_after_attempt, before_sleep_log, retry_if_exception, wait_exponential
from typing import Final
from requests.exceptions import Timeout, HTTPError, ConnectionError, RequestException
from .dassana_exception import ApiRequest, ApiResponse, ApiError, NetworkError, ServerError, RateLimitError, AuthError
from requests.compat import (
    basestring,
    urlencode
)
from requests.utils import to_key_val_list
from json import dumps as json_dumps

logger: Final = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def encode_params(data):
    """Encode parameters in a piece of data.

    Will successfully encode parameters when passed as a dict or a list of
    2-tuples. Order is retained if data is a list of 2-tuples but arbitrary
    if parameters are supplied as a dict.
    """

    if data is None:
        return None
    if isinstance(data, (str, bytes)):
        return data
    elif hasattr(data, "read"):
        return data
    elif hasattr(data, "__iter__"):
        result = []
        for k, vs in to_key_val_list(data):
            if isinstance(vs, basestring) or not hasattr(vs, "__iter__"):
                vs = [vs]
            for v in vs:
                if v is not None:
                    result.append(
                        (
                            k.encode("utf-8") if isinstance(k, str) else k,
                            v.encode("utf-8") if isinstance(v, str) else v,
                        )
                    )
        return urlencode(result, doseq=True)
    else:
        return data


def call_api(method, url, data=None, json=None, auth=None, headers=None, params=None, cookies=None, timeout=300,
             verify=True, is_internal=False, ignore_not_found_error=False, new_status_validator=None,
             do_not_track_request_body=False) -> Response:
    api_start_ts = timeit.default_timer()
    try:
        response = api_request(method, url, data, json, auth, headers, params, cookies, timeout, verify, is_internal,
                               ignore_not_found_error, new_status_validator, do_not_track_request_body)
        api_end_ts = timeit.default_timer()
        logging.debug(f"API request successful (url - {url} body - {data or json})")
        return response
    except ApiError as e:
        api_end_ts = timeit.default_timer()
        logging.error(f"{str(e)}")
        raise e


@retry(
    retry=retry_if_exception(lambda e: isinstance(e, ApiError) and e.is_auto_recoverable),
    wait=wait_exponential(multiplier=5, min=30, max=60),
    stop=stop_after_attempt(3),
    before_sleep=before_sleep_log(logger, logging.WARNING),
    reraise=True)
def api_request(method, url, data=None, json=None, auth=None, headers=None, params=None, cookies=None, timeout=300,
                verify=True, is_internal=False, ignore_not_found_error=False, new_status_validator=None,
                do_not_track_request_body=False) -> Response:
    global status_validator
    http_request = ApiRequest(method, url,
                              (json_dumps(json) if json is not None else encode_params(data))
                              if not do_not_track_request_body else None)
    try:
        response = requests.request(method, url, headers=headers, data=data, json=json, params=params, auth=auth,
                                    timeout=timeout, cookies=cookies, verify=verify)
        http_response = ApiResponse().from_response(response)
        status_validator = new_status_validator or status_validator
        status_validator(http_request, http_response, is_internal, ignore_not_found_error)
        return response
    except (ConnectionError, Timeout) as exp:
        raise NetworkError(http_request, exp, is_internal=is_internal)
    except HTTPError as httpError:
        raise ApiError(http_request, ApiResponse().from_response(httpError.response), is_internal=is_internal,
                       is_auto_recoverable=True)
    except RequestException as requestError:
        raise ApiError(http_request, ApiResponse().from_response(requestError.response), error_details=requestError,
                       is_internal=is_internal, is_auto_recoverable=True)
    except ApiError as apiError:
        raise apiError


def status_validator(http_request, http_response, is_internal, ignore_not_found_error):
    if int(http_response.status_code / 100) == 2:
        return
    elif http_response.status_code == 400:
        raise ApiError(http_request, http_response, is_internal=is_internal, is_auto_recoverable=False)
    elif not ignore_not_found_error and http_response.status_code == 404:
        raise ApiError(http_request, http_response, is_internal=is_internal, is_auto_recoverable=False)
    elif http_response.status_code in (401, 403):
        raise AuthError(http_request, http_response, is_internal)
    elif http_response.status_code == 408:
        raise NetworkError(http_request, http_response, is_internal=is_internal)
    elif http_response.status_code == 429:
        raise RateLimitError(http_request, http_response, is_internal=is_internal)
    elif int(http_response.status_code / 100) == 5:
        raise ServerError(http_request, http_response, is_internal=is_internal)
    else:
        raise ApiError(http_request, http_response, is_internal=is_internal, is_auto_recoverable=True)



import datetime
import gzip
import logging
import threading
import time
from typing import Final

import boto3
import requests
from google.cloud import storage

from .api import call_api
from .dassana_env import *
from .dassana_exception import *
from .dassana_logging import log

logger: Final = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

job_list = set()


def datetime_handler(val):
    if isinstance(val, datetime.datetime):
        return val.isoformat()
    return str(val)


def compress_file(file_name):
    with open(file_name, 'rb') as file_in:
        with gzip.open(f"{file_name}.gz", 'wb') as file_out:
            file_out.writelines(file_in)
    logger.info("Compressed file completed")


def get_headers():
    headers = {}
    if is_internal_auth():
        access_token = get_access_token()
        headers = {
            "x-dassana-tenant-id": get_tenant_id(),
            "Authorization": f"Bearer {access_token}",
        }
    else:
        app_token = get_dassana_token()
        headers = {
            "Authorization": f"Dassana {app_token}"
        }
    return headers


def get_exc_str(exc):
    return str(exc).replace("\"", "").replace("'", "").replace("\n", " ").replace("\t", " ")


def patch_ingestion(job_id, metadata=None):
    if metadata is None:
        metadata = {}
    json_body = {"metadata": metadata}
    res = call_api('PATCH', get_ingestion_srv_url() + "/job/" + job_id, headers=get_headers(), json=json_body,
                   is_internal=True)
    return res.json()


def every(delay, task):
    next_time = time.time() + delay
    while True:
        time.sleep(max(0, next_time - time.time()))
        task()
        next_time += (time.time() - next_time) // delay * delay + delay


def iterate_and_update_job_status():
    global job_list
    for job_id in job_list:
        try:
            patch_ingestion(job_id)
            logger.info(f"Updated job status for job id: {job_id}")
        except Exception as exp:
            logger.warning(f"Failed to update job ({job_id}) heartbeat due to {exp}")


threading.Thread(target=lambda: every(1800, iterate_and_update_job_status), daemon=True).start()


def get_ingestion_config(ingestion_config_id, app_id):
    app_url = get_app_url()
    url = f"https://{app_url}/app/{app_id}/ingestionConfig/{ingestion_config_id}"
    headers = get_headers()
    response = call_api("GET", url, headers=headers,
                        verify=False if app_url.endswith("svc.cluster.local:443") else True,
                        is_internal=True)
    
    try:
        response = response.json()
        if not response.get("config"):
            raise KeyError("config  missing in the response.")
        else:
            if not response["config"].get("_selectedScopeIds"):
                raise KeyError("_selectedScopeIds are missing in the response.")            
        return response
    except:
        raise

def patch_ingestion_config(ingestion_config_id, app_id, payload):
    app_url = get_app_url()
    url = f"https://{app_url}/app/{app_id}/ingestionConfig/{ingestion_config_id}"
    headers = get_headers()
    response = call_api("PATCH", url=url, headers=headers, json=payload,
                        verify=False if app_url.endswith("svc.cluster.local:443") else True,
                        is_internal=True)
    return response.text

def get_access_token():
    auth_url = get_auth_url()
    url = f"{auth_url}/oauth/token"
    data = {
        "grant_type": "client_credentials",
        "client_id": get_client_id(),
        "client_secret": get_client_secret(),
    }
    response = call_api("POST", url, data=data, verify=False if auth_url.endswith("svc.cluster.local:443") else True,
                        is_internal=True)
    return response.json()["access_token"]

class DassanaWriter:
    def __init__(self, source, record_type, config_id, metadata=None, priority=None, is_snapshot=False,
                 file_size_limit=249):
        if metadata is None:
            metadata = {}
        logger.info("Initialized common utility")

        self.file_size_limit = file_size_limit
        self.source = source
        self.record_type = record_type
        self.config_id = config_id
        self.metadata = metadata
        self.priority = priority
        self.is_snapshot = is_snapshot
        self.bytes_written = 0
        self.fail_counter = 0
        self.pass_counter = 0
        self.debug_log = set()
        self.storage_service = None
        self.client = None
        self.aws_iam_role_arn = None
        self.aws_iam_external_id = None
        self.aws_sts_client = None
        self.aws_session_token_expiration = None
        self.bucket_name = None
        self.blob = None
        self.full_file_path = None
        self.ingestion_service_url = get_ingestion_srv_url()
        self.is_internal_auth = is_internal_auth()
        self.file_path = self.get_file_path()
        self.job_id = None
        self.ingestion_metadata = None
        self.custom_file_dict = dict()
        self.initialize_client()
        log(scope_id=self.metadata["scope"]["scopeId"], job_id=self.job_id)
        self.file = open(self.file_path, 'a')

    def get_file_path(self):
        epoch_ts = int(time.time())
        if not self.is_internal_auth:
            return f"/tmp/{epoch_ts}.ndjson"
        return f"{epoch_ts}.ndjson"

    def initialize_client(self):
        global job_list
        response = self.get_ingestion_details()
        if "jobId" not in response or "stageDetails" not in response or "cloud" not in response["stageDetails"]:
            raise InternalError("Invalid job created with missing details")

        self.storage_service = response['stageDetails']['cloud']
        self.job_id = response["jobId"]
        logger.info(f"Ingestion job created with job id: {self.job_id}")
        if "metadata" in response:
            self.ingestion_metadata = response["metadata"]
        else:
            self.ingestion_metadata = {}
        if "creationTs" in response:
            self.ingestion_metadata["creationTs"] = response["creationTs"]
        else:
            self.ingestion_metadata["creationTs"] = int(time.time() * 1000)
        job_list.add(self.job_id)

        if "bucket" in response['stageDetails']:
            self.bucket_name = response['stageDetails']['bucket']
            self.full_file_path = response['stageDetails']['filePath']

        if self.storage_service == 'gcp':
            if "bucket" in response["stageDetails"]:
                credentials = response['stageDetails']['serviceAccountCredentialsJson']
                with open('service_account.json', 'w') as f:
                    json.dump(json.loads(credentials), f, indent=4)
                    f.close()

                os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = 'service_account.json'
                self.client = storage.Client()
        elif self.storage_service == 'aws':
            stage_details = response['stageDetails']
            if "awsIamRoleArn" in stage_details and stage_details["awsIamRoleArn"] is not None:
                self.aws_sts_client = boto3.client('sts', aws_access_key_id=stage_details['accessKey'],
                                                   aws_secret_access_key=stage_details['secretKey'])
                self.aws_iam_role_arn = stage_details['awsIamRoleArn']
                self.aws_iam_external_id = stage_details['awsIamExternalId']
            else:
                self.client = boto3.client('s3', aws_access_key_id=stage_details['accessKey'],
                                           aws_secret_access_key=stage_details['secretKey'])

    def write_json(self, json_object):
        self.file.flush()
        json.dump(json_object, self.file)
        self.file.write('\n')
        self.bytes_written = self.file.tell()
        if self.bytes_written >= int(self.file_size_limit) * 1000 * 1000:
            self.file.close()
            compress_file(self.file_path)
            self.upload_to_cloud(self.file_path)
            self.file_path = self.get_file_path()
            self.file = open(self.file_path, 'a')
            logger.info(f"Ingested data: {self.bytes_written} bytes")
            self.bytes_written = 0

    def write_custom_json(self, json_object, file_name):
        if file_name in self.custom_file_dict:
            custom_file = self.custom_file_dict[file_name]
        else:
            custom_file = open(file_name, 'a')
            self.custom_file_dict[file_name] = custom_file
        custom_file.flush()
        json.dump(json_object, custom_file)
        custom_file.write('\n')

    def upload_to_cloud(self, file_name):
        try:
            if not self.is_internal_auth:
                self.upload_to_signed_url()
            elif self.storage_service == 'gcp':
                self.upload_to_gcp(file_name)
            elif self.storage_service == 'aws':
                self.upload_to_aws(file_name)
            else:
                raise StageWriteFailure("Unsupported stage")
        except Exception as exp:
            raise StageWriteFailure(str(exp))

        if os.path.exists(file_name):
            os.remove(file_name)
        if os.path.exists(file_name + ".gz"):
            os.remove(file_name + ".gz")

    def upload_to_gcp(self, file_name):
        if self.client is None:
            raise ValueError("GCP client not initialized.")

        self.blob = self.client.bucket(self.bucket_name).blob(str(self.full_file_path) + "/" + str(file_name) + ".gz")
        self.blob.upload_from_filename(file_name + ".gz")

    def upload_to_aws(self, file_name):
        if self.client is None and self.aws_sts_client is None:
            raise ValueError("AWS client not initialized")

        if self.aws_iam_role_arn and (not self.aws_session_token_expiration or (
                self.aws_session_token_expiration.timestamp() < (
                datetime.datetime.now() + datetime.timedelta(minutes=2)).timestamp())):
            assume_role_response = self.aws_sts_client.assume_role(
                RoleArn=self.aws_iam_role_arn,
                RoleSessionName="DassanaIngestion",
                ExternalId=self.aws_iam_external_id)
            temp_credentials = assume_role_response['Credentials']
            self.aws_session_token_expiration = temp_credentials['Expiration']
            self.client = boto3.client(
                's3',
                aws_access_key_id=temp_credentials['AccessKeyId'],
                aws_secret_access_key=temp_credentials['SecretAccessKey'],
                aws_session_token=temp_credentials['SessionToken'])

        self.client.put_object(Body=open(f"{file_name}.gz", 'rb'), Bucket=self.bucket_name,
                               Key=f"{str(self.full_file_path)}/{str(file_name)}.gz")

    def upload_to_signed_url(self):
        signed_url = self.get_signing_url()
        if not signed_url:
            raise ValueError("The signed URL has not been received")

        headers = {
            'Content-Encoding': 'gzip',
            'Content-Type': 'application/octet-stream'
        }
        with open(str(self.file_path) + ".gz", "rb") as read:
            data = read.read()
            requests.put(url=signed_url, data=data, headers=headers)
            
    def cancel_job_with_error_info(self, error_code, failure_reason, fail_type="failed", error_message=None, is_internal=True, is_auto_recoverable=False):
        if not error_message:
            error_message = "Unexpected error occurred while collecting data"

        global job_list
        job_list.discard(self.job_id)
        if os.path.exists("service_account.json"):
            os.remove("service_account.json")
        metadata = {}
        fail_type_status_metadata = "canceled" if str(fail_type) == "cancel" else str(fail_type)
        self.debug_log.add(get_exc_str(str(failure_reason)))
        job_result = {"status": fail_type_status_metadata,
                      "debug_log": list(self.debug_log),
                      "pass": self.pass_counter, "fail": self.fail_counter,
                      "error_code": error_code,
                      "error_message": error_message,
                      "is_internal": is_internal,
                      "is_auto_recoverable": is_auto_recoverable}
        metadata["job_result"] = job_result
        self.cancel_ingestion_job(metadata, fail_type)
        log(status=fail_type, scope_id=self.metadata["scope"]["scopeId"], metadata=job_result)

    def cancel_job(self, exception_from_src):
        global job_list
        job_list.discard(self.job_id)
        if os.path.exists("service_account.json"):
            os.remove("service_account.json")
        str_exc = get_exc_str(str(exception_from_src))
        self.debug_log.add(str_exc)
        job_result_metadata = dict()
        job_result_metadata["status"] = "failed"
        job_result_metadata["pass"] = self.pass_counter
        job_result_metadata["fail"] = self.fail_counter
        job_result_metadata["debug_log"] = list(self.debug_log)

        if isinstance(exception_from_src, DassanaException):
            job_result_metadata["error_code"] = exception_from_src.error_type
            job_result_metadata["is_internal"] = exception_from_src.is_internal
            job_result_metadata["is_auto_recoverable"] = exception_from_src.is_auto_recoverable
            if exception_from_src.error_type == "internal_error":
                job_result_metadata["error_message"] = exception_from_src.message
            if isinstance(exception_from_src, ApiError):
                job_result_metadata["api_details"] = dict()
                job_result_metadata["api_details"]["request"] = exception_from_src.http_request.__dict__
                job_result_metadata["api_details"]["response"] = exception_from_src.http_response.__dict__
        else:
            job_result_metadata["error_code"] = "internal_error"
            job_result_metadata["error_message"] = "Unexpected error occurred while collecting data"
            job_result_metadata["is_internal"] = True
            job_result_metadata["is_auto_recoverable"] = False

        metadata = {"job_result": job_result_metadata}
        self.cancel_ingestion_job(metadata, "failed") 
        log(status=job_result_metadata["status"], scope_id=self.metadata["scope"]["scopeId"], exception=exception_from_src, metadata=job_result_metadata)

    def close(self, metadata=None):
        if metadata is None:
            metadata = {}
        global job_list
        job_list.discard(self.job_id)
        self.file.close()
        job_result = {"status": "ready_for_loading",
                      "source": {"pass": int(self.pass_counter), "fail": int(self.fail_counter),
                                 "debug_log": list(self.debug_log)}}
        metadata["job_result"] = job_result
        if self.bytes_written > 0:
            compress_file(self.file_path)
            self.upload_to_cloud(self.file_path)
            logger.info(f"Ingested remaining data: {self.bytes_written} bytes")
            self.bytes_written = 0
        for custom_file in self.custom_file_dict:
            self.custom_file_dict[custom_file].close()
            compress_file(custom_file)
            self.upload_to_cloud(custom_file)
        if os.path.exists("service_account.json"):
            os.remove("service_account.json")
        self.update_ingestion_to_done(metadata)
        log(status=job_result["status"], scope_id=self.metadata["scope"]["scopeId"],  metadata=job_result,job_id=self.job_id)

    def update_ingestion_to_done(self, metadata):
        json_body = {
            "metadata": metadata
        }
        res = call_api("POST", self.ingestion_service_url + "/job/" + self.job_id + "/" + "done", headers=get_headers(),
                       json=json_body, is_internal=True)
        logger.debug(f"Response Status: {res.status_code}")
        logger.debug(f"Request Body: {res.request.body}")
        logger.debug(f"Response Body: {res.text}")
        return res.json()

    def get_ingestion_details(self):
        json_body = {
            "source": str(self.source),
            "recordType": str(self.record_type),
            "configId": str(self.config_id),
            "is_snapshot": self.is_snapshot,
            "priority": self.priority,
            "metadata": self.metadata
        }
        if json_body["priority"] is None:
            del json_body["priority"]

        res = call_api("POST", self.ingestion_service_url + "/job/", headers=get_headers(), json=json_body,
                       is_internal=True)
        return res.json()

    def cancel_ingestion_job(self, metadata, fail_type):
        json_body = {
            "metadata": metadata
        }
        res = call_api("POST", self.ingestion_service_url + "/job/" + self.job_id + "/" + fail_type,
                       headers=get_headers(), json=json_body, is_internal=True)
        logger.debug(f"Response Status: {res.status_code}")
        logger.debug(f"Request Body: {res.request.body}")
        logger.debug(f"Response Body: {res.text}")
        return res.json()

    def get_signing_url(self):
        res = call_api("GET", self.ingestion_service_url + "/job/" + self.job_id + "/" + "signing-url",
                       headers=get_headers(), is_internal=True)
        signed_url = res.json()["url"]
        return signed_url
